



if not exist %LIBRARY_BIN%\faiss_avx2.dll exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_LIB%\faiss_avx2.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\AutoTune.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\Clustering.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IVFlib.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\Index.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\Index2Layer.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexBinary.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexBinaryFlat.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexBinaryFromFloat.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexBinaryHNSW.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexBinaryHash.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexBinaryIVF.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexFlat.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexHNSW.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexIVF.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexIVFFlat.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexIVFPQ.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexIVFPQFastScan.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexIVFPQR.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexIVFSpectralHash.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexLSH.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexLattice.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexPQ.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexPQFastScan.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexPreTransform.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexRefine.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexReplicas.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexScalarQuantizer.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\IndexShards.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\MatrixStats.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\MetaIndexes.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\MetricType.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\VectorTransform.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\clone_index.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\index_factory.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\index_io.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\AuxIndexStructures.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\FaissAssert.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\FaissException.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\HNSW.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\PolysemousTraining.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\ProductQuantizer-inl.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\ProductQuantizer.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\ResultHandler.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\ScalarQuantizer.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\ThreadedIndex-inl.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\ThreadedIndex.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\io.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\io_macros.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\lattice_Zn.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\platform_macros.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\pq4_fast_scan.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\impl\simd_result_handlers.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\invlists\BlockInvertedLists.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\invlists\DirectMap.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\invlists\InvertedLists.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\invlists\InvertedListsIOHook.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\AlignedTable.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\Heap.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\WorkerThread.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\distances.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\extra_distances.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\hamming-inl.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\hamming.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\ordered_key_value.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\partitioning.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\quantize_lut.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\random.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\simdlib.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\simdlib_avx2.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\simdlib_emulated.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\utils\utils.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuAutoTune.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuCloner.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuClonerOptions.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuDistance.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuFaissAssert.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndex.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndexBinaryFlat.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndexFlat.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndexIVF.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndexIVFFlat.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndexIVFPQ.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndexIVFScalarQuantizer.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuIndicesOptions.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\GpuResources.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\StandardGpuResources.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\BinaryDistance.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\BinaryFlatIndex.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\BroadcastSum.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\Distance.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\DistanceUtils.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\FlatIndex.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\GeneralDistance.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\GpuScalarQuantizer.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\IVFAppend.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\IVFBase.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\IVFFlat.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\IVFFlatScan.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\IVFInterleaved.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\IVFPQ.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\IVFUtils.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\InterleavedCodes.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\L2Norm.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\L2Select.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\PQCodeDistances-inl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\PQCodeDistances.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\PQCodeLoad.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\PQScanMultiPassNoPrecomputed-inl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\PQScanMultiPassNoPrecomputed.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\PQScanMultiPassPrecomputed.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\RemapIndices.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\impl\VectorResidual.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\BlockSelectKernel.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Comparators.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\ConversionOperators.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\CopyUtils.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\DeviceDefs.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\DeviceTensor-inl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\DeviceTensor.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\DeviceUtils.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\DeviceVector.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Float16.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\HostTensor-inl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\HostTensor.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Limits.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\LoadStoreOperators.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\MathOperators.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\MatrixMult-inl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\MatrixMult.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\MergeNetworkBlock.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\MergeNetworkUtils.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\MergeNetworkWarp.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\NoTypeTensor.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Pair.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\PtxUtils.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\ReductionOperators.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Reductions.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Select.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\StackDeviceMemory.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\StaticUtils.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Tensor-inl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Tensor.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\ThrustAllocator.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Timer.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\Transpose.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\WarpPackedBits.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\WarpSelectKernel.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\WarpShuffles.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\blockselect\BlockSelectImpl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\faiss_avx2\gpu\utils\warpselect\WarpSelectImpl.cuh exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
